import os
import logging
import numpy as np
import plotly.graph_objects as go
from dataclasses import dataclass, InitVar, field

log_dir = '/'.join(os.getcwd().split('/')[0:5]) + '/logs/'
logging.basicConfig(
     filename=f'{log_dir}datastores.log',
     format= '[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
     datefmt='%H:%M:%S'
)
logger = logging.getLogger('datastores')
logger.setLevel(logging.INFO)

SIG_MAP = {
    'PLETH': 'ppg',
    'ABP': 'abp',
}


@dataclass
class SignalStore:
    """Object to store wfdb.Record signal data and format."""

    data: np.ndarray
    fmt: str

    def replace_nan(self, value: float = 1.0):
        """Replace NaN values in the signal.

        Args:
            value (float): Replacement value. Defaults to 1.0.

        Raises:
            TypeError: Param value must be type float.
        """
        if not isinstance(value, float):
            raise TypeError('replace_nan: value must be type float')
        self.data[np.isnan(self.data)] = value

    def plot(self, mode: str = 'line', show: bool = True, **kwargs):
        """Plot signal data.

        Args:
            mode (str, optional): One of ['line', 'hist']. Defaults to 'line'.
            show (bool, optional): Show the plot. If False plot is returned. Defaults to True.
            **kwargs: Plotly graph parameters.

        Returns:
            fig (plotly.Figure): Plotly figure.
        """
        fig = go.Figure()
        if mode == 'line':
            fig.add_scatter(y=self.data, **kwargs)
        elif mode == 'hist':
            fig.add_histogram(x=self.data, **kwargs)
        if show:
            fig.show()
        else:
            return fig


@dataclass
class SignalGroup:
    """Object to store a group of SignalStore objects.
       Currently support and ppg, abp waveforms.
    """

    signals: InitVar[dict]
    ppg: SignalStore = field(init=False)
    abp: SignalStore = field(init=False)
    _names: list = field(init=False)

    def __post_init__(self, signals):
        names = []
        for sig, store in signals.items():
            setattr(self, SIG_MAP[sig], store)
            names.append(SIG_MAP[sig])
        setattr(self, '_names', names)

    def info(self):
        """Print signal names, formats and lengths."""
        out = ''
        for sig in self._names:
            store = getattr(self, sig)
            out += f'{sig: <7}: '
            out += f'Format {store.fmt}, Length of {store.data.shape[0]} samples\n'
        print(out)



import os
import typing
import numpy as np
from scipy import signal
from datetime import datetime

def get_iso_dt() -> None:
    """Get date in ISO string format.

    Returns:
        str: Date in the form YYYYMMDD.
    """
    return datetime.today().isoformat().split('T')[0].replace('-', '')

def build_data_directory(repo_dir: str, partner: str, date=None) -> str:
    """Create empty directory for new dataset.

    Args:
        repo_dir (str): Path to database_tools clone.
        partner (str): On of [mimic3, vital].

    Returns:
        data_dir (str): Path to data directory.
    """
    if date is None:
        date = get_iso_dt()
    data_dir = repo_dir + f'{partner}-data-{date}/'
    if os.path.exists(data_dir):
        return data_dir
    else:
        os.chdir(repo_dir)
        os.mkdir(data_dir)
        os.mkdir(data_dir + 'data/')
        os.mkdir(data_dir + 'data/lines')
        os.mkdir(data_dir + 'data/records')
        os.mkdir(data_dir + 'data/records/train')
        os.mkdir(data_dir + 'data/records/val')
        os.mkdir(data_dir + 'data/records/test')
        return data_dir

def window(x: np.ndarray, win_len: int, overlap: int) -> typing.List:
    """
    Gets indices for all windows in signal segment.

    Args:
        x (np.ndarray): Signal data.
        win_len (int): Length of windows (including overlap).
        overlap (int): Length of overlap between windows.

    Returns:
        idx (List): List of indices for windows.
    """
    idx = [[0, win_len]]
    idx = idx + [
        [(i * win_len) - overlap,
         ((i + 1) * win_len) - overlap] for i in range(1, int(len(x) / win_len))
    ]
    return idx

def make_equal_len(x: np.ndarray, y: np.ndarray) -> typing.Tuple[np.ndarray]:
    """Make two 1D numpy arrays equal length by padding (with 0s) the
       one that is longer.

    Args:
        x (np.ndarray): 1D array.
        y (np.ndarray): 1D array.

    Returns:
        x, y typing.Tuple[np.ndarray]: Tuple of arrays with one now padded.
    """
    len_x = len(x)
    len_y = len(y)
    if len_x > len_y:
        y = np.pad(y, pad_width=[0, len_x - len_y])
    else:
        x = np.pad(x, pad_width=[0, len_y - len_x])
    return (x, y)

def resample_signal(sig: np.ndarray, fs_old: int, fs_new: int) -> typing.Tuple[list, int]:
    """Resample a signal to a new sampling rate. This is done with the context
       of a reference length of time in order to produce a result that is
       evenly divisible by the window length (at the new sampling rate).

    Args:
        sig (list): Data.
        fs_old (int): Old sampling rate.
        fs_new (int): New sampling rate.

    Returns:
        resamp (list): Resampled signal.
    """
    frame_len = sig.reshape(-1).shape[0]
    frame_time = frame_len / fs_old
    resamp = signal.resample(sig, int(round(frame_time * fs_new, -1)))
    return resamp

def repair_peaks_troughs_idx(peaks: np.ndarray, troughs: np.ndarray) -> typing.Tuple[list, list]:
    """Takes a list of peaks and troughs and removes
       out of order elements. Regardless of which occurs first,
       a peak or a trough, a peak must be followed by a trough
       and vice versa.

    Args:
        peaks (list): Signal peaks.
        troughs (list): Signal troughs.

    Returns:
        first_repaired (list): Input with out of order items removed.
        second_repaired (list): Input with out of order items removed.

        Items are always returned with peaks idx as first tuple item.
    """
    # Configure algorithm to start with lowest index.
    try:
        if peaks[0] < troughs[0]:
            first = peaks
            second = troughs
        else:
            second = peaks
            first = troughs
    except IndexError:
        return (np.array([]), np.array([]))

    first_repaired, second_repaired = [], []  # lists to store outputs
    i_first, i_second = 0, 0  # declare starting indices
    for _ in enumerate(first):
        try:
            poi_1 = first[i_first]
            poi_2 = second[i_second]
            if poi_1 < poi_2:  # first point of interest is before second
                poi_3 = first[i_first + 1]
                if poi_2 < poi_3:  # second point of interest is before third
                    first_repaired.append(poi_1)
                    second_repaired.append(poi_2)
                    i_first += 1
                    i_second += 1
                else:  # first without iterating second
                    i_first += 1
            else: # inverse of other else condition
                i_second += 1
        except IndexError: # catch index error (always thrown in last iteration)
            first_repaired.append(poi_1)
            second_repaired.append(poi_2)

    # remove duplicates
    first_repaired = sorted(list(set(first_repaired)))
    second_repaired = sorted(list(set(second_repaired)))

    # place indices in the correct order
    try:
        if peaks[0] < troughs[0]:
            return (np.array(first_repaired), np.array(second_repaired))
        else:
            return (np.array(second_repaired), np.array(first_repaired))
    except IndexError:
        return (np.array([]), np.array([]))



import glob
import numpy as np
import pickle as pkl
from tqdm import tqdm
import tensorflow as tf
from time import time_ns
from typing import Tuple, List
from sklearn.model_selection import train_test_split
from database_tools.tools.dataset import Dataset

def generate_records(
    ds: Dataset,
    data_dir: str,
    split_strategy: Tuple[float, float, float] = (0.7, 0.15, 0.15),
    samples_per_file: int = 10000,
    scaler_path: str = None,
) -> Tuple[dict, dict]:
    """Generates TFRecords files from dataset.

    Args:
        ds (Dataset): Dataset object consisting of ppg and abp data from JSONLINES files.
                      Other data includes vpg, and apg data derived from ppg.

        split_strategy (Tuple[float, float, float], optional): Train, val, test split percentages.
            If not provided the entire dataset will be treated as test data.

        samples_per_file (int, optional): Max samples per tfrecords file. Defaults to 10000.

        scaler_path (path, optional): Path to an existing scaler. If not provided a scaler will be
            created based on given dataset.
    """
    print('Splitting data...')
    if split_strategy is None:
        idx = dict(train=[], val=[], test=[i for i in range(0, ds.ppg.shape[0])])
    else:
        idx = get_split_idx(n=ds.ppg.shape[0], split_strategy=split_strategy)
    data_unscaled = split_data(ds, idx)

    print('Scaling data...')
    if scaler_path is not None:
        with open(scaler_path, 'rb') as f:
            scaler, scaler_split_idx = pkl.load(f)
    else:
        scaler = None
    data_scaled, scaler_dict = scale_data(data_unscaled, scaler)

    print('Generating TFRecords...')
    write_records(
        data=data_scaled,
        data_dir=data_dir,
        samples_per_file=samples_per_file,
    )
    if scaler_path is None:
        with open(f'{data_dir}records_info_{time_ns()}.pkl', 'wb') as f:
            pkl.dump([scaler_dict, idx], f)
    return (data_unscaled, data_scaled, scaler_dict)

def get_split_idx(n: int, split_strategy: Tuple[float, float, float]) -> dict:
    idx = [i for i in range(0, n)]
    tmp = split_strategy[1] + split_strategy[2]  # val + test size
    idx_train, idx_val = train_test_split(
        idx,
        train_size=split_strategy[0],
        test_size=tmp
    )
    idx_val, idx_test = train_test_split(
        idx_val,
        train_size=round(split_strategy[1] / tmp, 2),
        test_size=round(split_strategy[2] / tmp, 2)
    )
    return dict(train=idx_train, val=idx_val, test=idx_test)

def split_data(ds: Dataset, idx) -> dict:
    data = {}
    for key in ['ppg', 'vpg', 'apg', 'abp']:
        tmp = ds.__getattribute__(key)
        data[key] = dict(
            train=tmp[idx['train']],
            val=tmp[idx['val']],
            test=tmp[idx['test']],
        )
    return data

def scale_data(data_unscaled: dict, scaler: dict) -> Tuple[dict, dict]:
    data_scaled = {'ppg': {}, 'vpg': {}, 'apg': {}, 'abp': {}}
    scaler_dict = {}
    for key in ['ppg', 'vpg', 'apg']:
        if scaler is None:
            min_ = np.min(data_unscaled[key]['train'])
            max_ = np.max(data_unscaled[key]['train'])
        else:
            min_ = scaler[key][0]
            max_ = scaler[key][1]
        scaler_dict[key] = [min_, max_]

        for split in ['train', 'val', 'test']:
            tmp = data_unscaled[key][split]
            tmp_scaled = np.divide(tmp - min_, max_ - min_)
            data_scaled[key][split] = tmp_scaled
    data_scaled['abp']['train'] = data_unscaled['abp']['train']
    data_scaled['abp']['val'] = data_unscaled['abp']['val']
    data_scaled['abp']['test'] = data_unscaled['abp']['test']
    return (data_scaled, scaler_dict)

def write_records(data: dict, data_dir: str, samples_per_file: int) -> None:
    records_dir = data_dir + 'data/records/'
    for split in ['train', 'val', 'test']:
        print(f'Starting {split} split...')
        split_data = {f'{sig}': data[sig][split] for sig in data.keys()}

        file_number = 0
        num_samples = 0
        examples = []
        total = split_data['ppg'].shape[0]
        for win in tqdm(zip(*split_data.values()), total=int(total)):
            examples.append(
                full_wave_window_example(
                    win=win,
                    labels=['ppg', 'vpg', 'apg', 'abp'],
                )
            )
            num_samples += 1
            if ((num_samples % samples_per_file) == 0) | (num_samples == total):
                file_name = records_dir + split + f'/mimic3_{str(file_number).zfill(3)}.tfrecords'
                with tf.io.TFRecordWriter(file_name) as w:
                    for tf_example in examples:
                        w.write(tf_example.SerializeToString())
                file_number += 1
                examples = []

def float_array_feature(value):
    """Returns a float_list from a float list."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=value))

def full_wave_window_example(win, labels):
    feature = {f'{l}': float_array_feature(win[i]) for i, l in enumerate(labels)}
    example = tf.train.Example(features=tf.train.Features(feature=feature))
    return example

def read_records(data_dir: str, n_cores: int = 10, n_parallel: int = tf.data.AUTOTUNE) -> dict:
    data = {}
    for split in ['train', 'val', 'test']:
        print(f'Reading {split} split.')
        filenames = [file for file in glob.glob(f'{data_dir}records/{split}/*.tfrecords')]
        dataset = tf.data.TFRecordDataset(
            filenames=filenames,
            compression_type=None,
            buffer_size=100000000,
            num_parallel_reads=n_cores
        )
        data[split] = dataset.map(full_wave_parse_window_function, num_parallel_calls=n_parallel)
    return data

def full_wave_parse_window_function(example_proto):
    features = tf.io.parse_single_example(
        example_proto, 
        features = {
            'ppg': tf.io.FixedLenFeature([256], tf.float32),
            'vpg': tf.io.FixedLenFeature([256], tf.float32),
            'apg': tf.io.FixedLenFeature([256], tf.float32),
            'abp': tf.io.FixedLenFeature([256], tf.float32),
        }
    )
    ppg = tf.reshape(features['ppg'], (256, 1))
    vpg = tf.reshape(features['vpg'], (256, 1))
    apg = tf.reshape(features['apg'], (256, 1))
    inputs = dict(ppg=ppg, vpg=vpg, apg=apg)
    label = tf.reshape(features['abp'], (256, 1))
    return inputs, label

def rescale_data(sig: np.ndarray, scaler: List[int]) -> np.ndarray:
    """Rescale sample(s) with min max scaler provided in the
       form of List[min, max].

    Args:
        sig (np.ndarray): Array of samples.
        scaler (dict): List of min and max value (min-max scaler).

    Returns:
        np.ndarray: Scaled sample(s).
    """
    sig_scaled = np.multiply(sig, scaler[1] - scaler[0]) + scaler[0]
    return sig_scaled



import os
import io
import wfdb
import random
import logging
import requests
import numpy as np
import pandas as pd
from typing import Union, List
from alive_progress import alive_bar

logging.basicConfig(
     filename=f'{__name__}.log',
     format= '[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
     datefmt='%H:%M:%S'
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

MIMIC_DIR = 'mimic3wdb/1.0/'

def generate_record_paths(name: str = None, shuffle: bool = True) -> str:
    if name is None:
        rec_dir = MIMIC_DIR + 'RECORDS'
    elif name == 'adults':
        rec_dir = MIMIC_DIR + 'RECORDS-adults'
    elif name == 'neonates':
        rec_dir = MIMIC_DIR + 'RECORDS-neonates'
    r = requests.get('https://physionet.org/files/' + rec_dir, stream=True)
    records = list(pd.read_csv(io.BytesIO(r.content), names=['records'])['records'])
    if shuffle:
        random.shuffle(records)
    logger.info(f'Successfuly got records file from {rec_dir}')
    for path in records:
        yield path[:-1]  # remove trailing /

def get_header_record(path: str, record_type: str) -> Union[wfdb.Record, wfdb.MultiRecord]:
    """Get data header or layout header record from MIMIC-III Waveforms database.

    Args:
        path (str): Path to data file.
        record_type (str): One of ['layout', 'data'].

    Returns:
        rec (wfdb.io.record.Record): WFDB record object.
    """
    pn_dir = MIMIC_DIR + path
    if record_type == 'layout':
        hea_name = path.split('/')[-1] + '_layout'
    elif record_type == 'data':
        pn_dir = pn_dir.rsplit('/', 1)[0] if '_' in pn_dir else pn_dir  # remove file name from path if there
        hea_name = path.split('/')[-1]
    else:
        raise ValueError('record_type must be one of [\'layout\', \'data\']')

    try:
        hea = wfdb.rdheader(pn_dir=pn_dir, record_name=hea_name)
        logger.info(f'Successfully got {record_type} record {hea_name}')
        return hea
    except Exception as e:
        logger.info(f'Failed to get {record_type} record {path} due to {e}')

def header_has_signals(hea: wfdb.Record, signals: List[str]) -> bool:
    """Check if a header record contains specified signals.

    Args:
        hea (wfdb.Record): Header file as wfdb.Record object.
        signals (List[str]): List of signals.

    Returns:
        bool: True if the record contains ALL signals provided.
    """
    try:
        if set(signals).difference(set(hea.sig_name)) == set():
            logger.info(f'Record {hea.record_name}.hea contains {signals}')
            return True
        else:
            logger.info(f'Record {hea.record_name}.hea does not contain {signals}')
            return False
    except TypeError:
        logger.info(f'Error getting {signals} from {hea.record_name}.hea')
        return False

def get_data_record(path: str, record_type: str = 'waveforms') -> wfdb.Record:
    """Get data record from MIMIC-III Waveforms database.

    Args:
        path (str): Path to data file.
        record_type (str): One of ['waveforms', 'numerics']. Defaults to 'numerics'.

    Returns:
        rec (wfdb.io.record.Record): WFDB record object.
    """
    if record_type == 'waveforms':
        pn_dir = MIMIC_DIR + '/'.join(path.split('/')[:-1])
        rcd_name = path.split('/')[-1]
    elif record_type == 'numerics':
        pn_dir = MIMIC_DIR + path
        rcd_name = path.split('/')[-1] + 'n'
    else:
        raise ValueError('record_type must be one of [\'waveforms\', \'numerics\']')
    try:
        rcd = wfdb.rdrecord(pn_dir=pn_dir, record_name=rcd_name)
        logger.info(f'Successfully got {record_type} record {rcd_name}')
        return rcd
    except Exception as e:
        logger.info(f'Failed to get {record_type} record {path} due to {e}')

def get_signal(rec: wfdb.Record, sig: str, errors: str = 'ignore') -> np.ndarray:
    """Extract signal data from a record.

    Args:
        rec (wfdb.Record): MIMIC-III record.
        sig (str): Signal name. One of ['PLETH', 'ABP'].
        errors (str, optional): One of ['raise', 'ignore'].

    Returns:
        data (np.ndarray): 1D signal data or NoneType if errors is 'ignore'.
    """
    try:
        s = rec.sig_name  # list of signals in record
        data = rec.p_signal[:, s.index(sig)].astype(np.float64)
        logger.info(f'Successfully extracted {sig} from {rec.record_name}')
        return data
    except Exception as e:
        if errors == 'ignore':
            logger.info(f'Failed to extract {sig} from {rec.record_name} due to {e}')
        elif errors == 'raise':
            raise ValueError(f'Signal name {sig} is not in the provided record')
        else:
            raise ValueError('errors must be one of [\'ignore\', \'raise\']')

def locate_valid_records(signals: List[str], min_length: int, n_segments: int, shuffle: bool = True) -> List[str]:
    """Locate valid data records. Exclusion is performed based on a list of signals
    the records must contain and a minimum length of the signals.

    Args:
        signals (List[str]): One or more of ['PLETH', 'ABP', ...]  TODO: add other signals
        min_length (int): Minimum length of data records to be considered valid.
        n_segments (int): Maximum number of records to find.
        shuffle (bool): If True records list is shuffled.

    Returns:
        valid_records (List[str]): List of valid segments.
    """
    valid_records = []
    with alive_bar(total=n_segments, bar='brackets', force_tty=True) as bar:
        for path in generate_record_paths(name='adults', shuffle=shuffle):

            # get patient layout header
            layout = get_header_record(path=path, record_type='layout')
            if layout is None: continue  # fx returns None if file DNE

            # check if header has provided signals
            if not header_has_signals(layout, signals): continue

            # get patient master header
            master = get_header_record(path=path, record_type='data')
            if master is None: continue

            # zip segment names and lengths
            for seg_name, n_samples in zip(master.seg_name, master.seg_len):
            
                # check segment length
                if (n_samples > min_length) & (seg_name != '~'):  # '~' indicates data is missing
                    seg_path = path + '/' + seg_name

                    # Get segment header
                    hea = get_header_record(path=seg_path, record_type='data')
                    if hea is None: continue

                    # Check if segment has provided signals and append
                    if header_has_signals(hea, signals):
                        valid_records.append(seg_path)
                        if n_segments is not None:
                            if len(valid_records) > n_segments:
                                return valid_records
                        bar()  # iterate loading bar



import numpy as np
import pandas as pd
import plotly.graph_objects as go

pd.options.plotting.backend = 'plotly'

class DataEvaluator():
    def __init__(self, stats):
        self._stats = stats
        self._stats = self._stats.fillna(0)
        self._sim = np.concatenate(
            (
                np.array(self._stats['time_sim']).reshape(-1, 1),
                np.array(self._stats['spec_sim']).reshape(-1, 1),
            ),
            axis=1
        )
        self._snr = np.concatenate(
            (
                np.array(self._stats['ppg_snr']).reshape(-1, 1),
                np.array(self._stats['abp_snr']).reshape(-1, 1),
            ),
            axis=1,
        )
        self._hr = np.concatenate(
            (
                np.array(self._stats['ppg_hr']).reshape(-1, 1),
                np.array(self._stats['abp_hr']).reshape(-1, 1),
            ),
            axis=1,
        )
        self._bp = np.concatenate(
            (
                np.array(self._stats['sbp']).reshape(-1, 1),
                np.array(self._stats['dbp']).reshape(-1, 1),
            ),
            axis=1,
        )
        self._beat_sim = np.concatenate(
            (
                np.array(self._stats['ppg_beat_sim']).reshape(-1, 1),
                np.array(self._stats['abp_beat_sim']).reshape(-1, 1),
            ),
            axis=1,
        )

    def run(self):
        figs = {}
        figs['sim'] = self.evaluate_sim()
        fig_p, fig_a = self.evaluate_snr()
        figs['ppg_snr'] = fig_p
        figs['abp_snr'] = fig_a
        figs['hr'] = self.evaluate_hr()
        fig_sbp, fig_dbp = self._evaluate_bp()
        figs['sbp'] = fig_sbp
        figs['dbp'] = fig_dbp
        fig_beat_sim_p, fig_beat_sim_a = self.evaluate_beat_sim()
        figs['ppg_beat_sim'] = fig_beat_sim_p
        figs['abp_beat_sim'] = fig_beat_sim_a
        return figs

    def evaluate_sim(self, bins=30):
        ppg, abp = self._sim[:, 0], self._sim[:, 1]
        fig = histogram3d(ppg, abp, range_=[[0, 1], [0, 1]], bins=bins)
        fig.update_layout(
                title={
                    'text': 'PPG, ABP Similarity Histogram',
                    'font': {'size': 35},
                },
                scene=dict(
                    xaxis={'title': 'Time Similarity (rₜ)', 'titlefont':{'size': 20}},
                    yaxis={'title': 'Spectral Similarity (r𝓌)', 'titlefont':{'size': 20}},
                    zaxis={'title': 'Number of Samples', 'titlefont':{'size': 20}},
                ),
                font={
                    'family': 'Courier New, monospace',
                    'color' : '#FFFFFF',
                    'size'  : 12,
                },
                template='plotly_dark',
        )
        fig.update_scenes(yaxis_autorange='reversed')
        return fig

    def evaluate_snr(self, bins=100):
        ppg_snr = pd.Series(self._snr[:, 0])
        abp_snr = pd.Series(self._snr[:, 1])
        ppg_snr[ppg_snr < -10] = -10
        ppg_snr[ppg_snr > 20] = 20
        abp_snr[abp_snr > 20] = 20
        abp_snr[abp_snr > 20] = 20 # modify extreme values for plotting

        fig_p  = ppg_snr.plot.hist(nbins=bins)
        fig_p.update_layout(
            title='PPG Signal-to-Noise Ratio Histogram',
            xaxis={'title': 'PPG SNR (dB)'},
            yaxis={'title': 'Number of Samples'},
            showlegend=False,
            template='plotly_dark',
        ) 
        fig_a = abp_snr.plot.hist(nbins=bins)
        fig_a.update_layout(
            title='ABP Signal-to-Noise Ratio Histogram',
            xaxis={'title': 'ABP SNR (dB)'},
            yaxis={'title': 'Number of Samples'},
            showlegend=False,
            template='plotly_dark',
        ) 
        return (fig_p, fig_a)

    def evaluate_beat_sim(self, bins=30):
        ppg = pd.Series(self._beat_sim[:, 0])
        abp = pd.Series(self._beat_sim[:, 1])
        fig_p  = ppg.plot.hist(nbins=bins)
        fig_a = abp.plot.hist(nbins=bins)
        return (fig_p, fig_a)

    def evaluate_hr(self, bins=25):
        ppg, abp = self._hr[:, 0], self._hr[:, 1]

        # Exclude extreme values
        ppg[ppg > 250] = 0
        abp[abp > 250] = 0

        min_ = np.min([np.min(ppg), np.min(abp)]) - 10
        max_ = np.max([np.max(ppg), np.max(abp)]) + 10
        fig = histogram3d(ppg, abp, range_=[[min_, max_], [min_, max_]], bins=bins)
        fig.update_layout(
                title={
                    'text': 'PPG, ABP HR Histogram',
                    'font': {'size': 35},
                },
                scene=dict(
                    xaxis={'title': 'PPG Heart Rate', 'titlefont':{'size': 20}},
                    yaxis={'title': 'ABP Heart Rate', 'titlefont':{'size': 20}},
                    zaxis={'title': 'Number of Samples', 'titlefont':{'size': 20}},
                ),
                font={
                    'family': 'Courier New, monospace',
                    'color' : '#FFFFFF',
                    'size'  : 12,
                },
                template='plotly_dark',
        )
        fig.update_scenes(
            yaxis_range=[max_, min_],
            xaxis_range=[min_, max_],
        )
        return fig

    def _evaluate_bp(self, bins=100):
        sbp = pd.Series(self._bp[:, 0])
        dbp = pd.Series(self._bp[:, 1])
        # sbp[sbp < -10] = -10
        # sbp[sbp > 20] = 20
        # dbp[dbp > 20] = 20
        # dbp[dbp > 20] = 20 # modify extreme values for plotting

        fig_sbp  = sbp.plot.hist(nbins=bins)
        fig_sbp.update_layout(
            title='Systolic Blood Pressure Histogram (max by window)',
            xaxis={'title': 'Blood Pressure (mmHg)'},
            yaxis={'title': 'Number of Samples'},
            showlegend=False,
            template='plotly_dark',
        ) 
        fig_dbp = dbp.plot.hist(nbins=bins)
        fig_dbp.update_layout(
            title='Diastolic Blood Pressure Histogram (min by window)',
            xaxis={'title': 'Blood Pressure (mmHg)'},
            yaxis={'title': 'Number of Samples'},
            showlegend=False,
            template='plotly_dark',
        ) 
        return (fig_sbp, fig_dbp)

def _bar_data(position3d, size=(1,1,1)):
    # position3d - 3-list or array of shape (3,) that represents the point of coords (x, y, 0), where a bar is placed
    # size = a 3-tuple whose elements are used to scale a unit cube to get a paralelipipedic bar
    # returns - an array of shape(8,3) representing the 8 vertices of  a bar at position3d
    
    bar = np.array([[0, 0, 0],
                    [1, 0, 0],
                    [1, 1, 0],
                    [0, 1, 0],
                    [0, 0, 1],
                    [1, 0, 1],
                    [1, 1, 1],
                    [0, 1, 1]], dtype=float) # the vertices of the unit cube

    bar *= np.asarray(size)# scale the cube to get the vertices of a parallelipipedic bar
    bar += np.asarray(position3d) #translate each  bar on the directio OP, with P=position3d
    return bar

def _triangulate_bar_faces(positions, sizes=None):
    # positions - array of shape (N, 3) that contains all positions in the plane z=0, where a histogram bar is placed 
    # sizes -  array of shape (N,3); each row represents the sizes to scale a unit cube to get a bar
    # returns the array of unique vertices, and the lists i, j, k to be used in instantiating the go.Mesh3d class

    if sizes is None:
        sizes = [(1,1,1)]*len(positions)
    else:
        if isinstance(sizes, (list, np.ndarray)) and len(sizes) != len(positions):
            raise ValueError('Your positions and sizes lists/arrays do not have the same length')
            
    all_bars = [_bar_data(pos, size)  for pos, size in zip(positions, sizes) if size[2]!=0]
    p, q, r = np.array(all_bars).shape

    # extract unique vertices from the list of all bar vertices
    vertices, ixr = np.unique(np.array(all_bars).reshape(p*q, r), return_inverse=True, axis=0)
    #for each bar, derive the sublists of indices i, j, k assocated to its chosen  triangulation
    I = []
    J = []
    K = []

    for k in range(len(all_bars)):
        I.extend(np.take(ixr, [8*k, 8*k+2,8*k, 8*k+5,8*k, 8*k+7, 8*k+5, 8*k+2, 8*k+3, 8*k+6, 8*k+7, 8*k+5])) 
        J.extend(np.take(ixr, [8*k+1, 8*k+3, 8*k+4, 8*k+1, 8*k+3, 8*k+4, 8*k+1, 8*k+6, 8*k+7, 8*k+2, 8*k+4, 8*k+6])) 
        K.extend(np.take(ixr, [8*k+2, 8*k, 8*k+5, 8*k, 8*k+7, 8*k, 8*k+2, 8*k+5, 8*k+6, 8*k+3, 8*k+5, 8*k+7]))  

    return  vertices, I, J, K  #triangulation vertices and I, J, K for mesh3d

def _get_plotly_mesh3d(x, y, range_, bins=5, bargap=0.05):
    bins = [bins, bins]
    # x, y- array-like of shape (n,), defining the x, and y-ccordinates of data set for which we plot a 3d hist
    hist, xedges, yedges = np.histogram2d(x, y,
                                          bins=bins,
                                          range=[range_[0],
                                                 range_[1]])
    xsize = xedges[1]-xedges[0]-bargap
    ysize = yedges[1]-yedges[0]-bargap
    xe, ye= np.meshgrid(xedges[:-1], yedges[:-1])
    ze = np.zeros(xe.shape)

    positions = np.dstack((xe, ye, ze))
    m, n, p = positions.shape
    positions = positions.reshape(m*n, p)
    sizes = np.array([(xsize, ysize, h) for h in hist.flatten()])
    vertices, I, J, K  = _triangulate_bar_faces(positions, sizes=sizes)
    X, Y, Z = vertices.T
    return X, Y, Z, I, J, K

def histogram3d(x, y, range_, bins=50):
    X, Y, Z, I, J, K = _get_plotly_mesh3d(x, y, range_, bins=bins, bargap=0)

    lighting = go.mesh3d.Lighting(
        ambient=0.6,
        roughness=0.1,
        specular=1.0,
        fresnel=5.0,
    )
    mesh3d = go.Mesh3d(
        x=X, y=Y, z=Z,
        i=I, j=J, k=K,
        colorscale='Viridis',
        intensity=np.linspace(0, 1, len(X), endpoint=True),
        flatshading=False,
        lighting=lighting,
    )
    layout = go.Layout(width=1000, 
                       height=1000, 
                       scene=dict(
                                  camera_eye_x=-1.0, 
                                  camera_eye_y=1.25,
                                  camera_eye_z=1.25,
                                 ),
                       yaxis={'autorange': False},
                       xaxis={},
                       font={},
                       )
    fig = go.Figure(data=[mesh3d], layout=layout)
    return fig